package scala.tools.nsc
package backend
package icode

import scala.tools.nsc.plugins
import java.io._
import plugins.Plugin
import tools.nsc._

/** The actual work is done here. */
abstract class GenCSharp extends Plugin with ThreeAddressCode
{
  /** The compiler.*/
  val global: Global

  import global._
  import icodes._

  def generateOutput()

  class GenCSharpPhase(prev: Phase) extends ICodePhase(prev)
  {
    def name = GenCSharpPlugin.PluginName

    override def run {
      if (settings.debug.value) inform("[running phase " + name + " on icode]")
      if (settings.Xdce.value)
        for ((sym, cls) <- icodes.classes ; if inliner.isClosureClass(sym) && !deadCode.liveClosures(sym))
          icodes.classes -= sym

      icodes.classes.values foreach apply

      generateOutput()
    }

    val tfa = new analysis.MethodTFA
    // the TypeKind on top of the stack is list.first 
    val preTypeStack = scala.collection.mutable.Map[(IMethod, BasicBlock, Int), List[TypeKind]]()
    val postTypeStack = scala.collection.mutable.Map[(IMethod, BasicBlock, Int), List[TypeKind]]()

    def populateTypeStacks(method: IMethod) {
      tfa.init(method)
      tfa.run
      for (currentBlock <- method.code.blocks) {
        var tfainfo = tfa.in(currentBlock)
        for ((ins, idx) <- currentBlock.toList.zipWithIndex) {
          val tfa.lattice.IState(entryBindings, typeStack: icodes.TypeStack) = tfainfo
          var listOfTypeKinds : List[TypeKind] = typeStack.types
          preTypeStack += ((method, currentBlock, idx) -> listOfTypeKinds)
          tfainfo = tfa.interpret(tfainfo, ins)
          val tfa.lattice.IState(entryBindings2, typeStack2) = tfainfo
          postTypeStack += ((method, currentBlock, idx) -> typeStack2.types)
        }
      }
    }

    val rdef = new icodes.reachingDefinitions.ReachingDefinitionsAnalysis
    // case class StackSlot(height: Int, mm: IMethod, bb: BasicBlock, idx: Int)
    // val depends = scala.collection.mutable.Map[(IMethod, BasicBlock, Int), Set[StackSlot]]()
    // val emptySetOfStackSlots = Set[StackSlot]()

    def displayDefsReachingStackSlots(method: IMethod) {
      scala.Console.println("method: " + method.toString)
      rdef.init(method)
      rdef.run
      for (currentBlock <- method.code.blocks) {
        scala.Console.println("\t block: " + currentBlock)
        /** Prints as a table the defs reaching instrIdx. */
        def makeSenseOfReachingDefs(rdefVars:  Set[(Local, BasicBlock, Int)],
                                    rdefStack: List[Set[(BasicBlock, Int)]],
                                    instrIdx:  Int) {
          //
          if(rdefStack.isEmpty){
             scala.Console.println("\t\t\t no reaching-defs on the empty stack")
          }
          for((slot, depth) <- rdefStack.zipWithIndex){
            scala.Console.println("\t\t\t reaching the slot at depth: " + depth)
            for((bb, bbidx) <- slot) {
              scala.Console.println("\t\t\t\t def: " + where(bb, bbidx))
            }
          }
          def where(b: BasicBlock, i: Int) = "/" + b(i) + " in block " + b + "\\"
        }

        var rdefinfo = rdef.in(currentBlock)
        for ((currentIns, currentIdx) <- currentBlock.toList.zipWithIndex) {
          // print typeStack and instruction each on its own line  
          scala.Console.println("\t\t\t type stack : " + preTypeStack(method, currentBlock, currentIdx).mkString("[", ",", "]"))
          // print reaching-defs as a table
          val icodes.reachingDefinitions.rdefLattice.IState(rdefVars,
                                                            rdefStack : List[Set[(BasicBlock, Int)]]) = rdefinfo
          makeSenseOfReachingDefs(rdefVars, rdefStack, currentIdx)
          scala.Console.println("\t\t" + currentIdx + "|\t" + currentIns)
          /** computes reaching-defs for the program point after idx. */
          rdefinfo = rdef.interpret(currentBlock, currentIdx, rdefinfo)
        }
        scala.Console.println("\t\t\t last type stack : " + postTypeStack(method, currentBlock, currentBlock.length-1).mkString("[", ",", "]"))
        val icodes.reachingDefinitions.rdefLattice.IState(rdefVars,
                                                          rdefStack : List[Set[(BasicBlock, Int)]]) = rdefinfo
        makeSenseOfReachingDefs(rdefVars, rdefStack, currentBlock.length - 1)
      }
    }

    def makeItLookGoodOnTheConsole(method: IMethod) {
      scala.Console.println("method: " + method.toString)
      for (currentBlock <- method.code.blocks) {
        scala.Console.println("\t block: " + currentBlock)
        for ((ins, idx) <- currentBlock.toList.zipWithIndex) {
          scala.Console.println("\t\t\t  type stack : " + preTypeStack(method, currentBlock, idx).mkString("[", ",", "]"))
          scala.Console.println("\t\t" + idx + "|\t" + ins)
        }
        scala.Console.println("\t\t\t  last type stack : " + postTypeStack(method, currentBlock, currentBlock.length-1).mkString("[", ",", "]"))
      }
    }

    /** Cut&pasted from GenICode.Context 
     * Make a fresh local variable. It ensures the 'name' is unique. */
    def makeLocal(unit: CompilationUnit, method: IMethod, pos: Position, tpe: Type, name: String): Local = {
      val sym = method.symbol.newVariable(pos, unit.fresh.newName(pos, name))
        .setInfo(tpe)
        .setFlag(scala.tools.nsc.symtab.Flags.SYNTHETIC)
      method.addLocal(new Local(sym, toTypeKind(tpe), false))
    }

    case class StackAccessingProgramPoint(ppbb: BasicBlock,
                                          ppidx: Int,
                                          ppDepthWithinInstr: Int /* set to 0 for all `pushing' instructions. */ 
                                         ) {
      override def toString() = { "/ in block " + ppbb + ", instruction " + ppbb(ppidx) + ", depth " + ppDepthWithinInstr + "\\" }
    }
    case class UseDefWeb(defs: scala.collection.mutable.Set[StackAccessingProgramPoint],
                         uses: scala.collection.mutable.Set[StackAccessingProgramPoint]) {
      def containsDef(bb: BasicBlock, idx: Int): Boolean = {
        defs contains StackAccessingProgramPoint(bb, idx, 0)
      }
      def containsUse(usebb: BasicBlock, useidx: Int, useDepth: Int) = {
        uses contains StackAccessingProgramPoint(usebb, useidx, useDepth)
      }
      def this(defbb: BasicBlock, defidx: Int, usebb: BasicBlock, useidx: Int, useDepth: Int) = {
        this(scala.collection.mutable.Set(StackAccessingProgramPoint(defbb, defidx, 0)),
             scala.collection.mutable.Set(StackAccessingProgramPoint(usebb, useidx, useDepth)))
      }
      def addUse(usebb: BasicBlock, useidx: Int, useDepth: Int) {
        uses += StackAccessingProgramPoint(usebb, useidx, useDepth)
      }
      def addDefinitions(defsToAdd: Set[(BasicBlock, Int)]) {
        val origsAsStackAccesses = defsToAdd map ( orig => StackAccessingProgramPoint(orig._1, orig._2, 0) )
        defs ++= origsAsStackAccesses
      }
      /** over all the types of uses in this web */
      def leastCommonAncestor(method: IMethod,
                              preTypeStack: scala.collection.Map[(IMethod, BasicBlock, Int), List[TypeKind]]) : TypeKind = {
        var ts = for (StackAccessingProgramPoint(ppbb, ppidx, ppDepth) <- uses)
                 yield (preTypeStack(method, ppbb, ppidx))(ppDepth)
        val t : TypeKind = ts reduceLeft icodes.lub
        t
      }
      def defsAsInstructions
        = defs map { case StackAccessingProgramPoint(ppbb, ppidx, ppDepth) => ppbb(ppidx) }
      /** used to avoid creating another local variable in case all definitions are LOAD_LOCALs from the same variable */
      def allDefsFromSameVar : Option[Local] = {
        assert (!(defs.isEmpty))
        val b = defsAsInstructions forall (ins => ins.isInstanceOf[opcodes.LOAD_LOCAL])
        if(!b) return None
        val loadedVars : scala.collection.Set[Local]
          = for (d <- defsAsInstructions; val lolo = d.asInstanceOf[opcodes.LOAD_LOCAL]) yield lolo.local 
        if (loadedVars.size == 1) Some(loadedVars.first)
        else None
      }
    }
    // val emptySetOfSMPP = Set[StackAccessingProgramPoint]()

    def computeWebs(cls: IClass, method: IMethod) : scala.collection.mutable.Set[UseDefWeb] = {
      rdef.init(method)
      rdef.run
      // use-def webs as in Sec. 8.10 of Muchnick's Advanced Compiler book
      val useDefWebs = new scala.collection.mutable.HashSet[UseDefWeb]() {
        def lookupForDef(bb: BasicBlock, idx: Int) : Option[UseDefWeb] = {
          val candidates = this filter (_ containsDef(bb, idx))
          assert (candidates.size <= 1)
          if(candidates.isEmpty) None
          else Some(candidates.first) 
        }
        def lookupForUse(bb: BasicBlock, idx: Int, depth: Int) : Option[UseDefWeb] = {
          val candidates = this filter (_ containsUse(bb, idx, depth))
          assert (candidates.size <= 1)
          if(candidates.isEmpty) None
          else Some(candidates.first)
        }
      }
      for (currentBlock <- method.code.blocks) {
        var rdefinfo = rdef.in(currentBlock)
        for ((currentInstruction, currentIdx) <- currentBlock.toList.zipWithIndex) {
          if (currentInstruction.consumed > 0) {
            // the info on reaching-defs for the current instruction
            val icodes.reachingDefinitions.rdefLattice.IState(rdefVars,
                                                              rdefStack : List[Set[(BasicBlock, Int)]]) = rdefinfo
            // for each popped value
            for (currentPoppedDepth <- 0 to (currentInstruction.consumed - 1) ) {
              /* the instruction(s) pushing that value will instead write to this var.
                 rdefStack contains for each stack depth the originating instructions.
                 focus only on those definitions for currentPoppedDepth.
                 In terms of the description in Muchnick's book, rdefStack keeps for each stack slot its ud-chain,
                 i.e., all the definitions that may flow to that use. Below, those definitions are
                 tracked in `originatingSet'
               */
              val originatingSet = rdefStack(currentPoppedDepth)

              for(originating <- originatingSet) {
                val webIfAny = useDefWebs.lookupForDef(originating._1, originating._2) 
                webIfAny match {
                  case None => {
                    /* the definition not yet in any web, but the use may be. If so, add the definitions to that web */
                    val webIfAny2 = useDefWebs.lookupForUse(currentBlock, currentIdx, currentPoppedDepth)
                    webIfAny2 match {
                      case None => {
                        /* neither the definition nor the use are in any webs,
                           create a new one (with the current use as single use so far)
                         */
                        useDefWebs += new UseDefWeb(originating._1, originating._2, currentBlock, currentIdx, currentPoppedDepth)
                      }
                      case Some(web2) => {
                        web2.addDefinitions(originatingSet)
                      }
                    }
                  }
                  case Some(web) => {
                    /* the definition was in  a web, update in place the set of uses in that web adding the current use,
                     * and add all other definitions in originatingSet to that web too.  */
                    web.addUse(currentBlock, currentIdx, currentPoppedDepth)
                    if(originatingSet.size > 1) {
                      web.addDefinitions(originatingSet)
                    }
                    /* TODO nothing else to do for currentPoppedDepth, skip the rest of the traversal over originatingSet */
                  }
                }
              }
            }
          }
          /* get the reaching definitions ready for the next instruction in this basic block. */
          rdefinfo = rdef.interpret(currentBlock, currentIdx, rdefinfo)
        }
      }
      useDefWebs
    }

    override def apply(cls: IClass) {
      for (method <- cls.methods if method.code != null) {
        populateTypeStacks(method)
        displayDefsReachingStackSlots(method)
        computeWebs(cls, method)
      }
      for (method <- cls.methods if method.code != null) {
        // makeItLookGoodOnTheConsole(method)
        val useDefWebs = computeWebs(cls, method)
        // useDefWebs2Console(method, useDefWebs)
        getA3Code(cls, method)
      }
    }

    case class A3Result(newLocals: List[Local],
                        stmts: List[ThreeAddressStmt],
                        ins2stm: Map[(BasicBlock, Int), ThreeAddressStmt])
    def getA3Code(cls: IClass, method: IMethod) {
      val useDefWebs = computeWebs(cls, method)
      for(web <- useDefWebs) {
        // val UseDefWeb(defs, uses) = web
        val web2local = new scala.collection.mutable.HashMap[UseDefWeb, Local]
        val local = web.allDefsFromSameVar match {
          case Some(local) => local
          case None => {
            // create new local 
            val tk = web.leastCommonAncestor(method, preTypeStack)
            val newLocal = makeLocal(cls.cunit, method, NoPosition, tk.toType, "stackslot")
            assert (newLocal.kind == tk, "passed a Type to makeLocal (although a TypeKind was available) and got another TypeKind")
            newLocal
          }
        }
        web2local += (web -> local)
      }
      // TODO return an A3Result value
    }

    def useDefWebs2Console(method: IMethod, useDefWebs: collection.Set[UseDefWeb]) {
      // a DSL for .dot files would be great here ...
      scala.Console.println("use-def-webs for method: " + method)
      scala.Console.println()
      for(web <- useDefWebs) {
        val UseDefWeb(defs, uses) = web
        scala.Console.println("\t defs:")
        defs foreach (d => scala.Console.println("\t\t\t" + d))
        // scala.Console.println(">>>>")
        scala.Console.println("\t uses:")
        uses foreach (u => scala.Console.println("\t\t\t" + u))
        scala.Console.println()
      }
    }

  }


  class CSharpTextPrinter(writer: PrintWriter, lin: Linearizer) extends TextPrinter(writer, lin) {

    override def printBlock(bb: BasicBlock) {
      print(bb.label)
      if (bb.loopHeader) print("[loop header]")
      print(": ");
      if (settings.debug.value) print("pred: " + bb.predecessors + " succs: " + bb.successors + " flags: " + bb.flagsString)
      indent; println
      // bb foreach printInstruction
      for ((ins, idx) <- bb.toList.zipWithIndex) {
        print(idx)
        print("|\t")
        val msgConsumedTypes = ins.consumedTypes
        val msgProducedTypes = ins match {
          // TODO case cm : opcodes.CALL_METHOD => toTypeKind(cm.method.tpe.resultType) 
          case _ => ins.producedTypes
        }
        println(ins.toString())
        println("\t\t, consumed: " + ins.consumed + ", consumed types: " + msgConsumedTypes)
        println("\t\t, produced: " + ins.produced + ", produced types: " + msgProducedTypes)
      }
      undent; println
    }

    override def printMethod(m: IMethod) {
      println("/*-------------------------------------*/")
      print("def "); print(m.symbol.name);
      print("("); printList(printParam)(m.params, ", "); print(")");
      print(": "); print(m.symbol.info.resultType)

      if (!m.isDeferred) {
        println(" {")
        val localsWithTypes = m.locals map {l => (l, l.kind)}
        println("locals: " + localsWithTypes.mkString("[", ", ", "]"))
        println("startBlock: " + m.code.startBlock)
        println("blocks: " + m.code.blocks.mkString("[", ",", "]"))
        println
        lin.linearize(m) foreach printBlock
        println("}")

        indent; println("Exception handlers: ")
        m.exh foreach printExceptionHandler

        undent; println
      } else
        println
    }

  }

}

// val newVar = makeLocal(cls.cunit, method, NoPosition, typeKindOfNewVar.toType, "stackslot")
// assert(newVar.kind == typeKindOfNewVar, "had to pass a Type to makeLocal (although a TypeKind was available)")

